/*
 * SPDX-FileCopyrightText: (C) 2016 Red Hat, Inc. (www.redhat.com)
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifndef E_MAIL_CONFIG_EWS_OFFLINE_OPTIONS_H
#define E_MAIL_CONFIG_EWS_OFFLINE_OPTIONS_H

#include <glib-object.h>

G_BEGIN_DECLS

void e_mail_config_ews_offline_options_type_register (GTypeModule *type_module);

G_END_DECLS

#endif /* E_MAIL_CONFIG_EWS_OFFLINE_OPTIONS_H */
