/*
 * SPDX-FileCopyrightText: (C) 2020 Red Hat (www.redhat.com)
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifndef E_MAIL_PARSER_EWS_MULTIPART_MIXED_H
#define E_MAIL_PARSER_EWS_MULTIPART_MIXED_H

#include <glib-object.h>

G_BEGIN_DECLS

void e_mail_parser_ews_multipart_mixed_type_register (GTypeModule *type_module);

G_END_DECLS

#endif /* E_MAIL_PARSER_EWS_MULTIPART_MIXED_H */
