
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef CAMEL_EWS_ENUMTYPES_H
#define CAMEL_EWS_ENUMTYPES_H

#include <glib-object.h>

G_BEGIN_DECLS

/* Enumerations from "camel-ews-enums.h" */
#define CAMEL_TYPE_EWS_STORE_OOO_ALERT_STATE	(camel_ews_store_ooo_alert_state_get_type())
GType camel_ews_store_ooo_alert_state_get_type	(void) G_GNUC_CONST;

G_END_DECLS

#endif /* CAMEL_EWS_ENUMTYPES_H */

/* Generated data ends here */

